/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "native.pmh"

namespace fe
{

#if FE_2DGL==FE_2D_X_GFX

#define FE_X_EVENTMASK	ExposureMask		| \
						EnterWindowMask		| \
						LeaveWindowMask		| \
						KeyPressMask		| \
						KeyReleaseMask		| \
						ButtonPressMask		| \
						ButtonReleaseMask	| \
						StructureNotifyMask

/*** alternates:    PointerMotionMask/ButtonMotionMask
					StructureNotifyMask/SubstructureNotifyMask */


Result NativeWindow::nativeStartup(void)
{
	//* in case something goes wrong
	FE_WINDOW_HANDLE default_handle={ NULL, NULL, 0 };
	setWindowHandle(default_handle);

	FEASSERT(m_spEventContextI.isValid());
	sp<NativeEventContext> spNativeEventContext=m_spEventContextI;
	FEASSERT(spNativeEventContext.isValid());
	Display *pDisplay=spNativeEventContext->displayHandle();

	m_spEventContextI->threadLock();
	int screen=DefaultScreen(pDisplay);

	Window parent=RootWindow(pDisplay,screen);
	if(!parent)
	{
		feLog("NativeWindow::nativeStartup"
				" RootWindow(pDisplay=%p,screen=%p)=NULL\n",
				pDisplay,screen);
		return e_undefinedFailure;
	}

	int m=0;
	int attributes[16];
	attributes[m++]=GLX_RGBA;
	attributes[m++]=GLX_RED_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_GREEN_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_BLUE_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_DEPTH_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_DOUBLEBUFFER;
	attributes[m]=None;

	XVisualInfo *pVisualInfo;
	if( !(pVisualInfo=glXChooseVisual(pDisplay,screen,attributes)) )
	{
		feLog("NativeWindow::nativeStartup "
					"compatible visual not available through GLX.\n"
					"  Using default visual.\n");

		int nitems;
		XVisualInfo vinfo;
		vinfo.depth=DefaultDepth(pDisplay,screen);
		vinfo.visual=DefaultVisual(pDisplay,screen);
		pVisualInfo=XGetVisualInfo(pDisplay,VisualNoMask,&vinfo,&nitems);
	}

	unsigned long value_mask=0;
	XSetWindowAttributes cwa;

	cwa.border_pixel=0;
	cwa.colormap=XCreateColormap(pDisplay,parent,pVisualInfo->visual,
			AllocNone);

	value_mask=CWColormap|CWBorderPixel;

	//* create the window
	Window win=XCreateWindow(pDisplay,parent,
							m_box[0],m_box[1],			// position
							width(m_box),height(m_box),	// size
							0,							// border width
							pVisualInfo->depth,
							InputOutput,
							pVisualInfo->visual,
							value_mask,&cwa);

	FE_WINDOW_HANDLE handle={pDisplay, pVisualInfo, win};
	setWindowHandle(handle);

	setPointerMotion(spNativeEventContext->pointerMotion());

	//* set client name
	char hostname[128];
	int fail=gethostname(hostname,128);
	FEASSERT(!fail);
	XTextProperty clientName;
	char *stringlist[1];
	stringlist[0]=hostname;
	if(!XStringListToTextProperty(stringlist,1,&clientName))
	{
		feX("Could not allocate space for client name");
	}
	XSetWMClientMachine(pDisplay,win,&clientName);
	XFree((void *)clientName.value);

	//* set pid
	pid_t pid=getpid();
	Atom atom = XInternAtom(pDisplay, "_NET_WM_PID", FALSE);
	FEASSERT(atom!=None);
	XChangeProperty(pDisplay,win,atom,XA_CARDINAL,32,
			PropModeReplace,(unsigned char *)(&pid), 1);

	//* create GLX context
	GLXContext shareList=NULL;
	Bool direct=True;
	m_glxContext=glXCreateContext(pDisplay,pVisualInfo,shareList,direct);
	if(m_glxContext==NULL)
		feX("NativeWindow::nativeStartup","glXCreateContext failed\n");
	makeCurrent();

	spNativeEventContext->notifyCreate(m_windowHandle);

	//* force window to update now
	XFlush(pDisplay);
//	XSync(pDisplay,FALSE);

	m_spEventContextI->threadUnlock();

	setCreated(TRUE);
	return e_ok;
}

Result NativeWindow::nativeShutdown(void)
{
	if(created())
	{
		m_spEventContextI->threadLock();
		glXDestroyContext(windowHandle()->m_pDisplay,m_glxContext);
		XDestroyWindow(windowHandle()->m_pDisplay,windowHandle()->m_window);
		m_spEventContextI->threadUnlock();

		XFree(windowHandle()->m_pVisualInfo);
	}
	setCreated(FALSE);
	return e_ok;
}

void NativeWindow::setPointerMotion(BWORD active)
{
	long mask=FE_X_EVENTMASK | (active? PointerMotionMask: ButtonMotionMask);

	//* select which events we want to receive
	m_spEventContextI->threadLock();
	XSelectInput(windowHandle()->m_pDisplay,windowHandle()->m_window,mask);
	m_spEventContextI->threadUnlock();
}

void NativeWindow::setTitle(const String &title)
{
	//* set window title
	XTextProperty windowName;
	char *titles[1];
	titles[0]=(char *)title.c_str();
	if(!XStringListToTextProperty(titles,1,&windowName))
		feLog("Could not allocate space for window title");

	m_spEventContextI->threadLock();
	XSetWMName(windowHandle()->m_pDisplay,windowHandle()->m_window,&windowName);
	m_spEventContextI->threadUnlock();

	XFree((void *)windowName.value);
}

Result NativeWindow::open(const String title)
{
	if(mapped())
	{
		return e_alreadyAvailable;
	}

	startup();

	setTitle(title);

	//* map to screen and raise
	m_spEventContextI->threadLock();
	XMapRaised(windowHandle()->m_pDisplay,windowHandle()->m_window);
	m_spEventContextI->threadUnlock();

	setMapped(TRUE);

	makeCurrent();
	return e_ok;
}

Result NativeWindow::close(void)
{
	m_spEventContextI->threadLock();
	XUnmapWindow(windowHandle()->m_pDisplay,windowHandle()->m_window);
	m_spEventContextI->threadUnlock();

	setMapped(FALSE);
	return e_ok;
}

void NativeWindow::makeCurrent(void)
{
	m_spEventContextI->threadLock();
	Bool success=glXMakeCurrent(windowHandle()->m_pDisplay,
			windowHandle()->m_window,m_glxContext);
	m_spEventContextI->threadUnlock();

	if(!success)
	{
		feLog("NativeWindow::makeCurrent glXMakeCurrent failed\n");
		feLog(" glGetError: \"%s\"\n",gluErrorString(glGetError()));
	}
}

void NativeWindow::releaseCurrent(void)
{
	m_spEventContextI->threadLock();
	Bool success=glXMakeCurrent(windowHandle()->m_pDisplay,None,NULL);
	m_spEventContextI->threadUnlock();

	if(!success)
	{
		feLog("NativeWindow::releaseCurrent glXMakeCurrent failed\n");
		feLog(" glGetError: \"%s\"\n",gluErrorString(glGetError()));
	}
}

void NativeWindow::swapBuffers(void)
{
	if(m_glxContext)
	{
		m_spEventContextI->threadLock();
		glXSwapBuffers(windowHandle()->m_pDisplay,windowHandle()->m_window);
		m_spEventContextI->threadUnlock();
	}
}

#endif

} /* namespace */
