/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "viewer.pmh"

#define	FE_QV_DRAW_ORIGIN		TRUE
#define	FE_QV_DRAW_INTEREST		FALSE
#define	FE_QV_DRAW_AUXILLARY	FALSE

namespace fe
{

QuickViewer::~QuickViewer(void)
{
#ifdef FE_CHAINSIGNALER_PROFILING
	Peeker peeker;
	m_spSignalerI->peek(peeker);

	feLog("~QuickViewer:\n%s\n",peeker.str().c_str());
#endif
}

void QuickViewer::initialize(void)
{
	if(!getControllerI().isValid())
		feX("QuickViewer::initialize","need a valid ControllerI");

	Result result=registry()->manage("fexNativeWindowDL");
	if(failure(result))
		feX("QuickViewer::run","need fexNativeWindowDL");
	result=registry()->manage("fexOpenGLDL");
	if(failure(result))
		feX("QuickViewer::run","need fexOpenGLDL");

	sp<ViewerI> spViewerI(this);
	m_spSignalerI=registry()->create("SignalerI");
	m_spWindowI=registry()->create("WindowI");

	m_spScope=registry()->master()->catalog()->catalogComponent(
			"Scope","WinScope");
	m_asViewer.bind(m_spScope);

	if(!m_spWindowI.isValid() || !m_spScope.isValid() ||
			!m_spSignalerI.isValid() || !spViewerI.isValid())
	{
		feX("QuickViewer::initialize","couldn't create components");
	}

	m_spBeatLayout=m_spScope->declare("beat");

	m_spRenderLayout=m_spScope->declare("render");
	m_asViewer.populate(m_spRenderLayout);

	//* NOTE EventContext doesn't care what signal it is
	m_spSignalerI->insert(m_spWindowI->getEventContextI(),m_spBeatLayout);

	//* we are signaling ourself, but in the guise of ObjectViewer
	m_spSignalerI->insert(spViewerI,m_spRenderLayout);

	//* if there is joystick support loaded, use it
	m_spJoyHandler=registry()->create("JoyI");
	if(m_spJoyHandler.isValid())
	{
		insertEventHandler(m_spJoyHandler);
	}

	const Color bgcolor(0.2f,0.2f,0.2f);
	m_spWindowI->setBackground(bgcolor);
}

void QuickViewer::open(void)
{
	if(m_opened)
	{
		return;
	}
	m_opened=TRUE;

	const BWORD verbose=(System::getVerbose()=="all");

	sp<DrawI> spDrawI;
	const String draw_hint=
			registry()->master()->catalog()->catalogOrDefault<String>(
			"hint:DrawI", "");
	if(draw_hint != "")
	{
		if(verbose)
		{
			feLog("QuickViewer::open using \"%s\"\n", draw_hint.c_str());
		}
		spDrawI=registry()->create(draw_hint);
	}
	if(!spDrawI.isValid())
	{
		if(verbose)
		{
			feLog("QuickViewer::open using threaded draw\n");
		}
		spDrawI=registry()->create("*.DrawThreaded");
	}
	if(!spDrawI.isValid())
	{
		if(verbose)
		{
			feLog("QuickViewer::open using any draw\n");
		}
		spDrawI=registry()->create("DrawI");
	}

	if(!spDrawI.isValid())
	{
		feX("QuickViewer::open","couldn't create DrawI");
	}

	m_spWindowI->open("QuickViewer");

	bind(spDrawI);

	//* TODO proper create (perhaps a reusable component)
	m_spInterestHandler=new InterestHandler(m_spScope,spDrawI,
			getControllerI());
	m_spInterestHandler->setName("InterestHandler");

	sp<SignalerI> spViewSignaler(this);
	spViewSignaler->insert(m_spInterestHandler,m_spRenderLayout);
}

void QuickViewer::bind(sp<DrawI> spDrawI)
{
	ObjectViewer::bind(spDrawI);

	sp<ViewerI> spViewerI(this);
	spViewerI->bind(m_spWindowI);

	sp<EventContextI> spEventContextI=m_spWindowI->getEventContextI();
	if(spEventContextI.isValid())
	{
		spDrawI->setFontBase(spEventContextI->fontBase());
	}
}

void QuickViewer::insertBeatHandler(sp<HandlerI> spHandlerI)
{
	m_spSignalerI->insert(spHandlerI,m_spBeatLayout);
}

void QuickViewer::insertEventHandler(sp<HandlerI> spEventHandler)
{
	sp<Layout> spLayout = m_spScope->lookupLayout(FE_EVENT_LAYOUT);
	m_spSignalerI->insert(spEventHandler, spLayout);
}

void QuickViewer::insertDrawHandler(sp<HandlerI> spHandlerI)
{
	sp<SignalerI> spViewSignaler(this);
	spViewSignaler->insert(spHandlerI,m_spRenderLayout);
}

void QuickViewer::run(U32 frames)
{
	open();

	Record beat=m_spScope->createRecord(m_spBeatLayout);
	Record underlay=m_spScope->createRecord(m_spRenderLayout);
	Record render=m_spScope->createRecord(m_spRenderLayout);
	Record overlay=m_spScope->createRecord(m_spRenderLayout);

	m_asViewer.viewer_layer(underlay)=0;
	m_asViewer.viewer_perspective(underlay)=0;
	m_asViewer.render_draw(underlay)=getDrawI();
	m_asViewer.timestep(underlay)=1.0;

	m_asViewer.viewer_layer(render)=1;
	m_asViewer.viewer_perspective(render)=1;
	m_asViewer.render_draw(render)=getDrawI();
	m_asViewer.timestep(render)=1.0;

	m_asViewer.viewer_layer(overlay)=2;
	m_asViewer.viewer_perspective(overlay)=0;
	m_asViewer.render_draw(overlay)=getDrawI();
	m_asViewer.timestep(overlay)=1.0;

	U32 count=0;
	while((!frames || count!=frames) && !getControllerI()->closeEvent())
	{
		m_asViewer.time(underlay)=count;
		m_asViewer.time(render)=count;
		m_asViewer.time(overlay)=count;

		m_spSignalerI->signal(beat);
		m_spSignalerI->signal(underlay);
		m_spSignalerI->signal(render);
		m_spSignalerI->signal(overlay);
		getDrawI()->flush();

		if(!getDrawI()->isDirect())
		{
			//* don't flood a remote connection
			milliSleep(100);
		}

		count++;
	}
}

void QuickViewer::InterestHandler::handle(Record& render)
{
	if(m_asViewer.viewer_perspective.check(render) &&
		m_asViewer.viewer_perspective(render))
	{
		const Real pixelSize=m_spDrawI->view()->pixelSize(
				m_spControllerI->interestMatrix().translation(),1.0f);

#if FE_QV_DRAW_AUXILLARY
		const Real pixelDesiredAux=32.0f;
		const Real radiusAux=pixelDesiredAux/pixelSize;
		const Color lime(0.5f,1.0f,0.5f);
		m_spDrawI->drawTransformedMarker(m_spControllerI->auxillaryMatrix(),
				radiusAux,lime);
#endif

#if FE_QV_DRAW_ORIGIN
		const Real pixelDesiredOrigin=32.0f;
		const Real radiusOrigin=pixelDesiredOrigin/pixelSize;
		SpatialTransform identity;
		setIdentity(identity);

		const Color black(0.0f,0.0f,0.0f);
		m_spDrawI->drawTransformedMarker(identity,
				radiusOrigin,black);
#endif

#if FE_QV_DRAW_INTEREST
		const Real pixelDesired=16.0f;
		const Real radius=pixelDesired/pixelSize;
		const Color lavender(1.0f,0.5f,1.0f);
		m_spDrawI->drawTransformedMarker(m_spControllerI->interestMatrix(),
				radius,lavender);
#endif
	}
}

} /* namespace */
